# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
# service type constants:

from networking_plumgrid.neutron.plugins.common import constants as \
    net_pg_const
from networking_plumgrid.neutron.plugins.common import exceptions as plum_excep
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as policy_exc
from oslo_log import log as logging
from oslo_utils import uuidutils


LOG = logging.getLogger(__name__)


def _retrieve_subnet_dict(self, port_db, context):
    """
    Helper function to retrieve the subnet dictionary
    """
    # Retrieve subnet information
    subnet_db = {}
    if len(port_db["fixed_ips"]) != 0:
        for subnet_index in range(0,
                                  len(port_db["fixed_ips"])):
            subnet_id = (port_db["fixed_ips"][subnet_index]
                         ["subnet_id"])
            subnet_db[subnet_index] = self._get_subnet(context, subnet_id)
    return subnet_db


def _check_floatingip_in_use(self, fp_id, ptag_db):
    """
    Helper function to check if Floating IP already exists and
    is associated with a SecurityGroup/PolicyTag
    """
    for entry in ptag_db:
        if fp_id == entry["floatingip_id"]:
            return True
    return False


def _check_floatingip_network(self, context, ptag_db):
    """
    Helper function to check Floating IP is from the specified
    External Network
    """
    floatingip_db = self.get_floatingip(context,
                                        ptag_db["policy_tag"]["floatingip_id"])
    router_db = self.get_routers(context,
                   filters={"tenant_id": [ptag_db["policy_tag"]["tenant_id"]]})
    floating_net_id = floatingip_db["floating_network_id"]

    if floatingip_db.get('port_id') is not None:
        raise policy_exc.FloatingIPAlreadyInUse(id=str(floatingip_db["id"]))

    external_gateway = []
    if "router_id" in ptag_db["policy_tag"] and \
       ptag_db["policy_tag"]["router_id"] is not None:
        external_gateway.append(ptag_db["policy_tag"]["router_id"])
    else:
        for router in router_db:
            if ("external_gateway_info" in router and
                router["external_gateway_info"] and
                router["external_gateway_info"]
                      ["network_id"] == floating_net_id):
                external_gateway.append(router["id"])
        if len(external_gateway) == 0:
            raise policy_exc.GatewayNotSet(id=str(floatingip_db["id"]))
        if len(external_gateway) > 1:
            raise policy_exc.MultipleExternalGatewaysFound(
                                           gateway_list=external_gateway)
    return (floatingip_db, external_gateway[0])


def _check_duplicate_ports_policy_service_create(self, ps_db):
    """
    Helper function to check if duplicate ingress, egress, bidirectional
    ports exists in create policy service JSON
    """
    port_list = []
    if "ingress_ports" in ps_db and ps_db["ingress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                         in ps_db["ingress_ports"]]))
        if _check_duplicate_ports_config(ps_db["ingress_ports"]):
            return True, "ingress ports"

    if "egress_ports" in ps_db and ps_db["egress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                         in ps_db["egress_ports"]]))
        if _check_duplicate_ports_config(ps_db["egress_ports"]):
            return True, "egress ports"

    if "bidirectional_ports" in ps_db and ps_db["bidirectional_ports"]:
        port_list.append(set([tuple(d.items()) for d
                         in ps_db["bidirectional_ports"]]))
        if _check_duplicate_ports_config(ps_db["bidirectional_ports"]):
            return True, "bidirectional ports"

    if _check_duplicate_ports_across_config(port_list):
        return True, "create parameters"

    return False, None


def _check_duplicate_ports_policy_service_update(self, ps_db):
    """
    Helper function to check if duplicate ingress, egress, bidirectional
    ports exists in update policy service JSON
    """
    port_list = []
    if "add_ingress_ports" in ps_db and ps_db["add_ingress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                         in ps_db["add_ingress_ports"]]))
        if _check_duplicate_ports_config(ps_db["add_ingress_ports"]):
            return True, "add ingress ports"

    if "add_egress_ports" in ps_db and ps_db["add_egress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                          in ps_db["add_egress_ports"]]))
        if _check_duplicate_ports_config(ps_db["add_egress_ports"]):
            return True, "add egress ports"

    if "add_bidirectional_ports" in ps_db and ps_db["add_bidirectional_ports"]:
        port_list.append(set([tuple(d.items()) for d
                          in ps_db["add_bidirectional_ports"]]))
        if _check_duplicate_ports_config(ps_db["add_bidirectional_ports"]):
            return True, "add bidirectional ports"

    if "remove_ingress_ports" in ps_db and ps_db["remove_ingress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                          in ps_db["remove_ingress_ports"]]))
        if _check_duplicate_ports_config(ps_db["remove_ingress_ports"]):
            return True, "remove ingress ports"

    if "remove_egress_ports" in ps_db and ps_db["remove_egress_ports"]:
        port_list.append(set([tuple(d.items()) for d
                          in ps_db["remove_egress_ports"]]))
        if _check_duplicate_ports_config(ps_db["remove_egress_ports"]):
            return True, "remove engress ports"

    if ("remove_bidirectional_ports" in ps_db
        and ps_db["remove_bidirectional_ports"]):
        port_list.append(set([tuple(d.items()) for d
                          in ps_db["remove_bidirectional_ports"]]))
        if _check_duplicate_ports_config(ps_db["remove_bidirectional_ports"]):
            return True, "remove bidirectional ports"

    if _check_duplicate_ports_across_config(port_list):
        return True, "update parameters"

    return False, None


def _check_duplicate_ports_config(port_db):
    if (len(port_db) > len([dict(t) for t in
        set([tuple(d.items()) for d in port_db])])):
        return True


def _check_duplicate_ports_across_config(list_port_db):
    for i in range(0, len(list_port_db)):
        for j in range(i + 1, len(list_port_db)):
            if (len(list_port_db[i].intersection(list_port_db[j])) > 0):
                return True


def _validate_port_owner(port_db):
    """
    Helper function to validate port is owned by a compute resource
    """
    if ("device_owner" in port_db and "binding:vif_type" in port_db and
        "compute" in port_db["device_owner"] and
        port_db["binding:vif_type"] == net_pg_const.BINDING_VIF_TYPE_IOVISOR):
        return True
    return False


def _check_policy_service_leg_mode(ps_obj, updated_ps_obj=None):
    """
    Helper function to check if bidirectional ports are
    specified with ingress or egress ports
    """
    if ("ingress_ports" in ps_obj and "egress_ports"
        in ps_obj and "bidirectional_ports" in ps_obj):
        if updated_ps_obj is None:
            if ((ps_obj["ingress_ports"] or ps_obj["egress_ports"])
                and ps_obj["bidirectional_ports"]):
                return True
        else:
            if ps_obj["ingress_ports"] or ps_obj["egress_ports"]:
                if ("add_bidirectional_ports" in updated_ps_obj or
                    "remove_bidirectional_ports" in updated_ps_obj):
                    return True
            elif ps_obj["bidirectional_ports"]:
                if ("add_ingress_ports" in updated_ps_obj or
                    "add_egress_ports" in updated_ps_obj or
                    "remove_ingress_ports" in updated_ps_obj or
                    "remove_egress_ports" in updated_ps_obj):
                    return True
            else:
                #Update call with any prior ports in policy service
                if ((("add_ingress_ports" in updated_ps_obj or
                      "add_ingress_ports" in updated_ps_obj)
                    or ("remove_ingress_ports" in updated_ps_obj or
                        "remove_egress_ports" in updated_ps_obj))
                    and ("add_bidirectional_ports" in updated_ps_obj or
                         "remove_bidirectional_ports" in updated_ps_obj)):
                    return True
    return False


def _is_security_group(context, ep_obj, ep_db, config):
    if config in ep_db and ep_db[config]:
        for epg in ep_db[config]:
            epg_id_list = ep_obj.get_endpoint_groups(context,
                                     filters={'id': [epg['id']]},
                                     fields=["id"])
            if len(epg_id_list) == 1:
                if ("is_security_group" in epg_id_list[0]
                    and epg_id_list[0]["is_security_group"]):
                    operation = "Endpoint association"
                    raise policy_exc.OperationNotAllowed(operation=operation,
                                                       id=epg_id_list[0]["id"])


def _validate_ep_config(ep_db):
    if ((ep_db["ip_mask"] and ep_db["ip_port"]) or
        (ep_db["ip_mask"]and ep_db["port_id"]) or
        (ep_db["ip_port"] and ep_db["port_id"])):
        raise policy_exc.MultipleAssociationForEndpoint()


def _check_duplicates_endpoint_config(ep_db):
    if "ep_groups" in ep_db and ep_db["ep_groups"]:
        if (len(ep_db["ep_groups"]) > len([dict(t) for t in
            set([tuple(d.items()) for d in ep_db["ep_groups"]])])):
            raise policy_exc.DuplicateEndpointGroup()
    if "add_endpoint_groups" in ep_db and ep_db["add_endpoint_groups"]:
        if (len(ep_db["add_endpoint_groups"]) > len([dict(t) for t in
            set([tuple(d.items()) for d in ep_db["add_endpoint_groups"]])])):
            raise policy_exc.DuplicateEndpointGroup()
    if "remove_endpoint_groups" in ep_db and ep_db["remove_endpoint_groups"]:
        if (len(ep_db["remove_endpoint_groups"]) > len([dict(t) for t in
            set([tuple(d.items()) for d in
            ep_db["remove_endpoint_groups"]])])):
            raise policy_exc.DuplicateEndpointGroup()

    if (("add_endpoint_groups" in ep_db and ep_db["add_endpoint_groups"]) and
        ("remove_endpoint_groups" in ep_db
         and ep_db["remove_endpoint_groups"])):
        if (len(set([tuple(d.items()) for d in ep_db["add_endpoint_groups"]]).
            intersection(set([tuple(d.items()) for d in
            ep_db["remove_endpoint_groups"]]))) > 0):
            raise policy_exc.DuplicateEndpointGroup()


def _process_policy_tag(self, context, epg_db):
    if "policy_tag_id" in epg_db and epg_db["policy_tag_id"]:
        if (not uuidutils.is_uuid_like(epg_db['policy_tag_id']) and
            epg_db['policy_tag_id'] is not None):
            ptag_list = self.get_policy_tags(context,
                                 filters={'name': [epg_db['policy_tag_id']]},
                                 fields=["id"])

            if len(ptag_list) == 1:
                epg_db['policy_tag_id'] = ptag_list[0]["id"]
            elif len(ptag_list) == 0:
                err_message = ("No policy tag"
                               " matches found for tag"
                               " '%s'" % epg_db['policy_tag_id'])
                raise plum_excep.PLUMgridException(err_msg=err_message)
            else:
                err_message = ("Multiple policy tag"
                               " matches found for tag"
                               " '%s', use an ID to be more"
                               " specific." % epg_db['policy_tag_id'])
                raise plum_excep.PLUMgridException(err_msg=err_message)
    return epg_db


def _process_epg_update(self, context, epg_db):
    ptag_db = {}
    if "policy_tag_id" in epg_db and epg_db["policy_tag_id"]:
        config = "policy_tag_id"
    elif "add_tag" in epg_db and epg_db["add_tag"]:
        config = "add_tag"
    elif "remove_tag" in epg_db and epg_db["remove_tag"]:
        config = "remove_tag"
    else:
        config = None
    if config in epg_db and epg_db[config]:
        if (not uuidutils.is_uuid_like(epg_db[config]) and
            epg_db[config] is not None):
            ptag_list = self.get_policy_tags(context,
                                 filters={'name': [epg_db[config]]},
                                 fields=["id"])

            if len(ptag_list) == 1:
                epg_db[config] = ptag_list[0]["id"]
                ptag_db = self.get_policy_tag(context,
                                              ptag_list[0]["id"])
            elif len(ptag_list) == 0:
                err_message = ("No policy tag"
                               " matches found for tag"
                               " '%s'" % epg_db[config])
                raise plum_excep.PLUMgridException(err_msg=err_message)
            else:
                err_message = ("Multiple policy tag"
                               " matches found for tag"
                               " '%s', use an ID to be more"
                               " specific." % epg_db[config])
                raise plum_excep.PLUMgridException(err_msg=err_message)
        elif (uuidutils.is_uuid_like(epg_db[config]) and
              epg_db[config] is not None):
            ptag_id = epg_db[config]
            ptag_db = self.get_policy_tag(context,
                                          ptag_id)
    return (epg_db, ptag_db)


def _check_policy_service_in_use(self, context, id):
    """
    Helper function to check policy service is in use by
    any policy rule
    """
    pr_list = self.get_policy_rules(context,
                              fields=["id", "action_target"])
    action_target_list = []
    for rule in pr_list:
        if _check_remote_action_target(rule):
            action_target_list.append((rule["action_target"].split(":"))[1])
        else:
            action_target_list.append(rule["action_target"])
    in_use_list = filter(lambda x: x == id, action_target_list)
    if len(in_use_list) > 0:
        raise policy_exc.PolicyServiceInUsePolicyRule(id=id)


def _recursive_delete_endpoints(self, context, id):
    """
    Helper function to delete endpoints when endpoint group
    is deleted
    """
    endpoint_list = self.get_endpoints(context,
                                       filters={'ep_groups': [{'id': id}]})
    for endpoint in endpoint_list:
        if ((len(endpoint["ep_groups"]) == 1) and
            endpoint["ep_groups"][0]["id"] == id):
            self.delete_endpoint(context, endpoint["id"])


def _check_remote_action_target(pr_db):
    """
    Helper function to check if action target for policy rule
    is for a tenant id
    """
    if "action_target" in pr_db and pr_db["action_target"]:
        action_target_db = pr_db["action_target"].split(":")
        if len(action_target_db) == 2:
            return True
        else:
            return False



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_plumgrid._i18n import _
from neutronclient.common import extension


class EndpointGroup(extension.NeutronClientExtension):
    resource = 'endpoint_group'
    resource_plural = 'endpoint_groups'
    path = 'endpoint-groups'
    object_path = '/%s' % path
    resource_path = '/%s/%%s' % path
    versions = ['2.0']


def args2body(self, parsed_args):
    try:
        if parsed_args.name:
            ep_grp_name = parsed_args.name
            body = {'endpoint_group': {'name': ep_grp_name}}
        else:
            body = {'endpoint_group': {}}
        if parsed_args.tag:
            body['endpoint_group']['policy_tag_id'] = parsed_args.tag
        if parsed_args.description:
            body['endpoint_group']['description'] = parsed_args.description
        return body
    except KeyError as err:
        raise Exception("KeyError: " + str(err))


class EndpointGroupCreate(extension.ClientExtensionCreate,
                          EndpointGroup):
    """Create a Endpoint Group."""

    shell_command = 'endpoint-group-create'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name', metavar='<ENDPOINT-GROUP-NAME>',
            help=_('Descriptive name for endpoint group.'))
        parser.add_argument('--description', dest='description',
                        help=_('Description of the Endpoint Group '
                               'being created.'))
        parser.add_argument('--tag', dest='tag',
                        help=_('policy-tag name/uuid '))

    def args2body(self, parsed_args):
        body = args2body(self, parsed_args)
        if parsed_args.tenant_id:
            (body['endpoint_group']
                 ['tenant_id']) = parsed_args.tenant_id
        return body


class EndpointGroupList(extension.ClientExtensionList,
                      EndpointGroup):
    """List endpoint groups that belong to a given tenant."""

    shell_command = 'endpoint-group-list'
    list_columns = ['id', 'name', 'policy_tag_id', 'description',
                    'is_security_group']
    pagination_support = True
    sorting_support = True


class EndpointGroupShow(extension.ClientExtensionShow,
                        EndpointGroup):
    """Show information of a given endpoint group."""

    shell_command = 'endpoint-group-show'


class EndpointGroupDelete(extension.ClientExtensionDelete,
                          EndpointGroup):
    """Delete a given endpoint group."""

    shell_command = 'endpoint-group-delete'


class EndpointGroupUpdate(extension.ClientExtensionUpdate,
                          EndpointGroup):
    """Update a given endpoint-group."""

    shell_command = 'endpoint-group-update'

    def add_known_arguments(self, parser):
        parser.add_argument(
            '--name', metavar='name',
            help=_('Descriptive name for Endpoint Group'))
        parser.add_argument(
            '--description', metavar='description',
            help=_('Description for Endpoint Group'))
        parser.add_argument(
            '--add-tag',
            metavar='add_tag',
            help=_('Policy Tag Name/ID to be associated with this Endpoint '
                   'Group.'))
        parser.add_argument(
            '--remove-tag',
            metavar='remove_tag',
            help=_('Policy Tag Name/ID to be remove from this Endpoint '
                   'Group.'))

    def args2body(self, parsed_args):
        try:
            if parsed_args.name:
                ep_name = parsed_args.name
                body = {'endpoint_group': {'name': ep_name}}
            else:
                body = {'endpoint_group': {}}
            if parsed_args.description:
                (body['endpoint_group'][
                 'description']) = parsed_args.description
            if parsed_args.add_tag:
                (body['endpoint_group'][
                 'add_tag']) = parsed_args.add_tag
            if parsed_args.remove_tag:
                (body['endpoint_group'][
                 'remove_tag']) = parsed_args.remove_tag
            return body
        except KeyError as err:
            raise Exception("KeyError: " + str(err))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import common_db_mixin
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class ExtLinkMixin(common_db_mixin.CommonDbMixin):
    def get_ext_links(self, context, ext_list, filters=None, fields=None,
             sorts=None, limit=None, marker=None, page_reverse=None):
        """
        Gets the list of all the existing links/connectors for a tenant
        """
        return self._make_ext_link_dict(ext_list)

    def _make_ext_link_dict(self, ext_list):
        return ext_list



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa
from sqlalchemy import orm

from neutron.db import model_base
from neutron.db import models_v2


class ProviderNetBinding(model_base.BASEV2):
    """Represents binding of provider network extension."""
    __tablename__ = 'pg_provider_net_bindings'

    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey('networks.id', ondelete="CASCADE"),
                           primary_key=True)
    network_type = sa.Column(sa.String(32))
    physical_network = sa.Column(sa.String(64))
    vlan_id = sa.Column(sa.Integer)

    network = orm.relationship(
        models_v2.Network,
        backref=orm.backref("pgnetbinding", lazy='joined',
                            uselist=False, cascade='delete'))



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import mock
from networking_plumgrid.neutron.tests.unit.cli import test_cli20
from networking_plumgrid.neutronclient.policy import (
    endpoint as ep)
from neutronclient import shell
import sys


class CLITestV20ExtensionEndpointJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20ExtensionEndpointJSON,
              self).setUp(plurals={'tags': 'tag'})

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("endpoint", ep)]
        return contrib

    def test_ext_cmd_loaded(self):
        """Tests endpoint commands loaded."""
        shell.NeutronShell('2.0')
        ext_cmd = {'endpoint-list':
                   ep.EndpointList,
                   'endpoint-create':
                   ep.EndpointCreate,
                   'endpoint-update':
                   ep.EndpointUpdate,
                   'endpoint-delete':
                   ep.EndpointDelete,
                   'endpoint-show':
                   ep.EndpointShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_create_endpoint_name(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        args = [name]
        position_names = ['name']
        position_values = [name]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_port(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        port = 'portid'
        args = [name, '--port', port]
        position_names = ['name', 'port_id']
        position_values = [name, port]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_ip_mask(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_mask = '0.0.0.0/0'
        args = [name, '--ip-mask', ip_mask]
        position_names = ['name', 'ip_mask']
        position_values = [name, ip_mask]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_ip_port(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_port = '0.0.0.0_80_8'
        args = [name, '--ip-port', 'ip=0.0.0.0,port=80,mask=8']
        position_names = ['name', 'ip_port']
        position_values = [name, ip_port]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_epg(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        epg = [{'id': 'epgid'}]
        args = [name, '--endpoint-group', 'id=epgid']
        position_names = ['name', 'ep_groups']
        position_values = [name, epg]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_port_epg(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        port = 'portid'
        epg = [{'id': 'epgid'}]
        args = [name, '--port', port, '--endpoint-group', 'id=epgid']
        position_names = ['name', 'port_id', 'ep_groups']
        position_values = [name, port, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_port_epg_name(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        port = 'portid'
        epg = [{'id': 'epgid'}]
        args = [name, '--port', port, '--endpoint-group', 'name=epgid']
        position_names = ['name', 'port_id', 'ep_groups']
        position_values = [name, port, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_ip_mask_epg(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_mask = '0.0.0.0/0'
        epg = [{'id': 'epgid'}]
        args = [name, '--ip-mask', ip_mask, '--endpoint-group', 'id=epgid']
        position_names = ['name', 'ip_mask', 'ep_groups']
        position_values = [name, ip_mask, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_ip_mask_epg_name(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_mask = '0.0.0.0/0'
        epg = [{'id': 'epgid'}]
        args = [name, '--ip-mask', ip_mask, '--endpoint-group', 'name=epgid']
        position_names = ['name', 'ip_mask', 'ep_groups']
        position_values = [name, ip_mask, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_ip_port_epg(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_port = '0.0.0.0_80_8'
        epg = [{'id': 'epgid'}]
        args = [name, '--ip-port', 'ip=0.0.0.0,port=80,mask=8',
                '--endpoint-group', 'id=epgid']
        position_names = ['name', 'ip_port', 'ep_groups']
        position_values = [name, ip_port, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_ip_port_epg_name(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_port = '0.0.0.0_80_8'
        epg = [{'id': 'epgid'}]
        args = [name, '--ip-port', 'ip=0.0.0.0,port=80,mask=8',
                '--endpoint-group', 'name=epgid']
        position_names = ['name', 'ip_port', 'ep_groups']
        position_values = [name, ip_port, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_mul_classfication(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        epg = [{'id': 'epgid'}]
        port = 'portid'
        ip_mask = '0.0.0.0/0'
        args = [name, '--port', port, '--ip-mask', ip_mask,
                '--endpoint-group', 'id=epgid']
        position_names = ['name', 'port_id', 'ip_mask', 'ep_groups']
        position_values = [name, port, ip_mask, epg]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_invalid_ip_port(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        ip_port = '0.0.0.0_80_8'
        epg = [{'id': 'epgid'}]
        args = [name, '--ip-port', 'ip=0.0.0.0,port=80',
                '--endpoint-group', 'name=epgid']
        position_names = ['name', 'ip_port', 'ep_groups']
        position_values = [name, ip_port, epg]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_update_endpoint_name(self):
        # Update endpoint: myid --name
        resource = 'endpoint'
        cmd = ep.EndpointUpdate(test_cli20.MyApp(
                                      sys.stdout), None)
        args = ['myid', '--name', 'myep']
        values = {'name': 'myep'}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_endpoint_add_epg(self):
        # Update endpoint: myid --add-endpoint-group
        resource = 'endpoint'
        cmd = ep.EndpointUpdate(test_cli20.MyApp(
                                      sys.stdout), None)
        args = ['myid', '--add-endpoint-group', 'id=epg']
        values = {'add_endpoint_groups': [{'id': 'epg'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_endpoint_remove_epg(self):
        # Update endpoint: myid --remove-endpoint-group
        resource = 'endpoint'
        cmd = ep.EndpointUpdate(test_cli20.MyApp(
                                      sys.stdout), None)
        args = ['myid', '--remove-endpoint-group', 'id=epg']
        values = {'remove_endpoint_groups': [{'id': 'epg'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_endpoint_add_epg_name(self):
        # Update endpoint: myid --add-endpoint-group
        resource = 'endpoint'
        cmd = ep.EndpointUpdate(test_cli20.MyApp(
                                      sys.stdout), None)
        args = ['myid', '--add-endpoint-group', 'name=epg']
        values = {'add_endpoint_groups': [{'id': 'epg'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_update_endpoint_remove_epg_name(self):
        # Update endpoint: myid --remove-endpoint-group
        resource = 'endpoint'
        cmd = ep.EndpointUpdate(test_cli20.MyApp(
                                      sys.stdout), None)
        args = ['myid', '--remove-endpoint-group', 'name=epg']
        values = {'remove_endpoint_groups': [{'id': 'epg'}]}
        self._test_update_resource(resource, cmd, 'myid',
                                   args, values)

    def test_list_endpoints(self):
        # list endpoints: -D.
        resources = "endpoints"
        cmd = ep.EndpointList(test_cli20.MyApp(
                              sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_list_endpoints_sort(self):
        # list endpoints:
        # --sort-key name --sort-key id --sort-key asc --sort-key desc
        resources = "endpoints"
        cmd = ep.EndpointList(test_cli20.MyApp(
                              sys.stdout), None)
        self._test_list_resources(resources, cmd,
                                  sort_key=["name", "id"],
                                  sort_dir=["asc", "desc"])

    def test_list_endpoints_limit(self):
        # list endpoints: -P.
        resources = "endpoints"
        cmd = ep.EndpointList(test_cli20.MyApp(
                              sys.stdout), None)
        self._test_list_resources(resources, cmd, page_size=1000)

    def test_delete_endpoint(self):
        # Delete endpoint: myid.
        resource = "endpoint"
        cmd = ep.EndpointDelete(test_cli20.MyApp(
                                sys.stdout), None)
        my_id = 'myid'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)

    def test_show_endpoint(self):
        # Show endpoint: --fields id.
        resource = 'endpoint'
        cmd = ep.EndpointShow(test_cli20.MyApp(
                              sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])

    def test_create_endpoint_label(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        label = 'mylabel'
        args = [name, '--label', label]
        position_names = ['name', 'label']
        position_values = [name, label]
        self.assertRaises(Exception, self._test_create_resource, resource,
                          cmd, name, 'myid', args,
                          position_names, position_values)

    def test_create_endpoint_label_epg(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        label = 'mylabel'
        epg = [{'id': 'epgid'}]
        args = [name, '--label', label, '--endpoint-group', 'id=epgid']
        position_names = ['name', 'label', 'ep_groups']
        position_values = [name, label, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)

    def test_create_endpoint_label_epg_name(self):
        # Create endpoint: ep1.
        resource = 'endpoint'
        cmd = ep.EndpointCreate(test_cli20.MyApp(
                                sys.stdout), None)
        name = 'ep1'
        label = 'mylabel'
        epg = [{'id': 'epgid'}]
        args = [name, '--label', label, '--endpoint-group', 'name=epgid']
        position_names = ['name', 'label', 'ep_groups']
        position_values = [name, label, epg]
        self._test_create_resource(resource, cmd, name, 'myid', args,
                                   position_names, position_values)



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from networking_plumgrid._i18n import _
from neutron.common import exceptions


class InvalidPortForEndpoint(exceptions.InvalidInput):
    message = _("Port %(id)s not associated with a compute resource. Cannot "
                "be part of endpoint.")


class PortNotFound(exceptions.InvalidInput):
    message = _("Port %(id)s not found.")


class InvalidPolicyTagType(exceptions.InvalidInput):
    message = _("Invalid Policy tag type %(type)s. Supported "
                "types are: 'fip', 'dot1q', 'nsh'")


class NoPolicyTagFound(exceptions.NotFound):
    message = _("Policy Tag with id %(id)s does not exist")


class NoPolicyTagAssociation(exceptions.NotFound):
    message = _("Policy Tag with id %(id)s is not associated with "
                "Endpoint Group: %(epg_id)s")


class PolicyTagAlreadyInUse(exceptions.InUse):
    message = _("Policy Tag '%(ptag_id)s' already in use by endpoint "
                "group: '%(epg_id)s'.")


class PolicyTagAlreadyInUseSG(exceptions.InUse):
    message = _("Policy Tag '%(ptag_id)s' already in use by security "
                "group: '%(sg_id)s'.")


class PolicyTagAlreadyInUseMultipleEPG(exceptions.InUse):
    message = _("Policy Tag '%(ptag_id)s' is in use by multiple "
                "endpoint groups.")


class PolicyTagAlreadyInUseMultipleSG(exceptions.InUse):
    message = _("Policy Tag '%(ptag_id)s' is in use by multiple "
                "security groups.")


class SGInUseWithPolicyTag(exceptions.InUse):
    message = _("Security Group '%(sg_id)s' is in use by policy tag "
                ": '%(ptag_id)s'.")


class InvalidDataProvidedEndpointGroup(exceptions.InvalidInput):
    message = _("Invalid data provided for Endpoint Group.")


class InvalidDataProvidedEndpoint(exceptions.InvalidInput):
    message = _("Invalid data provided for Endpoint.")


class FloatingIPAlreadyInUse(exceptions.InUse):
    message = _("Floating IP '%(id)s' is in use by a port/policy-tag")


class FloatingIPMismatch(exceptions.InvalidInput):
    message = _("Floating IP '%(id)s' does not match with External Network.")


class GatewayNotSet(exceptions.InvalidInput):
    message = _("Failed to associate Floating IP '%(id)s'. Gateway not "
                "set for Router.")


class MultipleExternalGatewaysFound(exceptions.InvalidInput):
    message = _("Multiple External Gateways found '%(gateway_list)s'. Please "
                "provide specific Router ID.")


class NoPolicyServiceFound(exceptions.NotFound):
    message = _("Policy Service with id %(id)s does not exist.")


class PortAlreadyInUsePolicyService(exceptions.InUse):
    message = _("Port '%(port)s' already in use by a policy service "
                "'%(ps)s'.")


class InvalidPolicyServiceConfiguration(exceptions.InvalidInput):
    message = _("Policy service only supports ['ingress', 'egress', "
                "'bidirectional'] ports.")


class UpdateParametersRequired(exceptions.InvalidInput):
    message = _("No update parameter specified atleast one needed.")


class PolicyServiceInUse(exceptions.InUse):
    message = _("Policy Service %(id)s %(reason)s with port(s): %(port)s")

    def __init__(self, **kwargs):
        if 'reason' not in kwargs:
            kwargs['reason'] = _("is in use")
        super(PolicyServiceInUse, self).__init__(**kwargs)


class PolicyServiceInUsePolicyRule(exceptions.InUse):
    message = _("Policy Service %(id)s %(reason)s with policy rule(s).")

    def __init__(self, **kwargs):
        if 'reason' not in kwargs:
            kwargs['reason'] = _("is in use")
        super(PolicyServiceInUsePolicyRule, self).__init__(**kwargs)


class NoPolicyRuleFound(exceptions.NotFound):
    message = _("Policy Rule with id %(id)s does not exist.")


class InvalidPolicyRuleConfig(exceptions.InvalidInput):
    message = _("Endpoint group specified as '%(epg)s' of policy rule does "
                "not exist.")


class InvalidFormatActionTarget(exceptions.InvalidInput):
    message = _("Invalid input format for action-target. Use the following "
                "format 'service-name/uuid' OR '<tenant_id:service-name/uuid'")


class NoActionTargetFound(exceptions.InvalidInput):
    message = _("Action target '%(at)s' specified for policy rule does "
                "not exist.")


class OperationNotAllowed(exceptions.InvalidInput):
    message = _("%(operation)s operation for endpoint group (%(id)s) is "
                "not allowed. Endpoint group (%(id)s) refers to a security "
                "group.")


class NoPolicyTagFoundEndpointGroup(exceptions.InvalidInput):
    message = _("Policy Tag not associated with endpoint group '%(epg)s'."
                "Unable to apply tag to policy rule.")


class DuplicatePortFound(exceptions.InvalidInput):
    message = _("Duplicate ports found in %(match)s of policy service.")


class InvalidDataProvidedPolicyService(exceptions.InvalidInput):
    message = _("Invalid data provided for Policy Service.")


class InvalidPortForPolicyService(exceptions.InvalidInput):
    message = _("Port %(id)s not associated with a compute resource. Cannot "
                "be part of policy service.")


class PortAlreadyInUseEndpoint(exceptions.InUse):
    message = _("Port '%(port)s' already in use by a endpoint "
                "'%(ep)s'.")


class InvalidPortConfigPolicyService(exceptions.InvalidInput):
    message = _("Bidirectional ports cannot be specified along with "
                "ingress or egress ports")


class NotSupportedPortRangeICMP(exceptions.InvalidInput):
    message = _("Port range for ICMP protocol is not supported.")


class UnsupportedProtocol(exceptions.InvalidInput):
    message = _("Protocol (%(protocol)s) not supported for endpoint policy. "
                "Supported protocols are ['any', 'tcp', 'udp', 'icmp'].")


class UnsupportedAction(exceptions.InvalidInput):
    message = _("Action (%(action)s) not supported for endpoint policy. "
                "Supported action(s) are ['copy', 'allow'].")


class MultipleAssociationForEndpoint(exceptions.InvalidInput):
    message = _("Multiple association criterion for endpoint specified."
                " Please specify only one criteria per endpoint.")


class DuplicateEndpointGroup(exceptions.InvalidInput):
    message = _("Duplicate endpoint groups found in endpoint config.")


class InvalidInputActionTarget(exceptions.InvalidInput):
    message = _("Action target is required by the following "
                "policy rule action(s) : ['copy']")


class EndpointGroupAlreadyInUse(exceptions.InUse):
    message = ("Endpoint group '%(epg)s' already in use by Endpoint '%(ep)s'")


class ProtocolNotSpecified(exceptions.InvalidInput):
    message = ("Must also specify protocol if port range is given")



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_plumgrid._i18n import _
from neutron.api import extensions
from neutron.api.v2 import base
from neutron.common import exceptions as nexceptions
from neutron import manager


TVD = 'transit_domain'
TVDS = 'transit-domains'


class NoTransitDomainFound(nexceptions.NotFound):
    message = _("Transit domain with id %(id)s does not exist")


class TransitDomainInUse(nexceptions.InUse):
    message = _("Transit domain with id %(id)s is in use.")


RESOURCE_ATTRIBUTE_MAP = {
    TVD: {
        'id': {
            'allow_post': False,
            'allow_put': False,
            'validate': {
                'type:uuid': None
            },
            'is_visible': True,
            'primary_key': True
        },
        'name': {
            'allow_post': True,
            'allow_put': True,
            'is_visible': True,
            'default': ''
        },
        'implicit': {
            'allow_post': False,
            'allow_put': False,
            'is_visible': False,
            'default': False,
            'validate': {
                 'type:boolean': None
            }
        },
        'tenant_id': {
            'allow_post': True,
            'allow_put': False,
            'is_visible': True
        }
    }
}


class Transitdomain(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Transit Domain"

    @classmethod
    def get_alias(cls):
        return "transit-domain"

    @classmethod
    def get_description(cls):
        return "Transit domain API"

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/transit_domains" \
               "/api/v2.0"

    @classmethod
    def get_updated(cls):
        return "2012-10-05T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        exts = []
        plugin = manager.NeutronManager.get_plugin()

        resource_name = TVD
        collection_name = TVDS
        params = RESOURCE_ATTRIBUTE_MAP.get(resource_name, dict())
        controller = base.create_resource(
            collection_name, resource_name, plugin, params)
        ex = extensions.ResourceExtension(collection_name,
                                          controller)
        exts.append(ex)
        return exts



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as p_excep
from networking_plumgrid.neutron.plugins.extensions import \
    policytag
from neutron.api.v2 import attributes
from neutron.db import common_db_mixin
from neutron.db.l3_db import FloatingIP
from neutron.db import model_base
from neutron.db import models_v2
from oslo_log import log as logging
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc

LOG = logging.getLogger(__name__)


class PolicyTag(model_base.BASEV2, models_v2.HasId,
                models_v2.HasTenant):
    """DB definition for PLUMgrid Policy Tag object"""

    __tablename__ = "pg_policy_tags"

    name = sa.Column(sa.String(attributes.NAME_MAX_LEN))
    tag_type = sa.Column(sa.Enum('fip', 'dot1q', 'nsh',
                                name='policy_tag_type'))
    floatingip_address = sa.Column(sa.String(255))
    tag_id = sa.Column(sa.String(255))
    tenant_id = sa.Column(sa.String(36))
    router_id = sa.Column(sa.String(36))
    floatingip_id = sa.Column(sa.String(36),
                            sa.ForeignKey("floatingips.id",
                            ondelete="CASCADE"),
                            nullable=True)

    fp = orm.relationship(FloatingIP,
              backref=orm.backref("fp_binding",
                                  lazy="joined", cascade="all,delete"),
  primaryjoin="FloatingIP.id==PolicyTag.floatingip_id")


class PolicyTagMixin(common_db_mixin.CommonDbMixin):
    def create_policy_tag(self, context, policy_tag):
        """
        Creates a policy tag with specified tag type
        Args:
             policy_tag:
                   JSON object with policy tag attributes
                   name : display name policy tag
                   tenant_id : tenant uuid
                   id : endpoint group uuid
                   type : type of policy tag (fip|dot1q|nsh)
                   tag_id : tag id in range xx-yy
        """
        ptag = policy_tag["policy_tag"]

        if "floatingip_id" in ptag and ptag["floatingip_id"]:
            self._check_floatingip_in_use(context, ptag["floatingip_id"])

        with context.session.begin(subtransactions=True):
            ptag_db = PolicyTag(tenant_id=ptag["tenant_id"],
                                name=ptag["name"],
                                tag_type=ptag["tag_type"],
                                floatingip_id=ptag["floatingip_id"],
                                floatingip_address=ptag["floating_ip_address"],
                                router_id=ptag["router_id"],
                                tag_id=ptag["tag_id"])
            context.session.add(ptag_db)
        return self._make_ptag_dict(ptag_db)

    def get_policy_tag(self, context, ptag_id, fields=None):
        """
        Gets an existing policy tag
        Args:
             policy_tag_id = uuid of the policy tag being requested
        """
        try:
            query = self._model_query(context, PolicyTag)
            ptag_db = query.filter_by(id=ptag_id).one()
        except exc.NoResultFound:
            raise policytag.NoPolicyTagFound(id=ptag_id)
        return self._make_ptag_dict(ptag_db, fields)

    def get_policy_tags(self, context, filters=None, fields=None,
             sorts=None, limit=None, marker=None, page_reverse=None):
        """
        Gets the list of all the existing policy tags
        """
        return self._get_collection(context, PolicyTag,
                                    self._make_ptag_dict, filters=filters,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker, fields=fields,
                                    page_reverse=page_reverse)

    def delete_policy_tag(self, context, ptag_id):
        """
        Deletes an existing Policy tag
        Args:
             epg_id = uuid of the policy tag being deleted
        """
        try:
            query = context.session.query(PolicyTag)
            ptag_db = query.filter_by(id=ptag_id).first()
        except exc.NoResultFound:
            raise policytag.NoPolicyTagFound(id=ptag_id)
        with context.session.begin(subtransactions=True):
            context.session.delete(ptag_db)

    def update_policy_tag(self, context, ptag_id, policy_tag):
        """
        Updates an existing policy tag
        Args:
             ptag_id:
                   uuid of the policy tag being updated
             policy_tag:
                   JSON with updated attributes of the policy tag
                   name : name of policy tag
                   tenant_id : tenant uuid
                   id : uuid of policy tag
        """
        ptag = policy_tag["policy_tag"]
        if not ptag:
            raise policytag.UpdateParametersRequired()

        with context.session.begin(subtransactions=True):
            try:
                query = self._model_query(context, PolicyTag)
                ptag_db = query.filter_by(id=ptag_id).one()
            except exc.NoResultFound:
                raise policytag.NoPolicyTagFound(id=ptag_id)
            if 'name' in ptag:
                ptag_db.update({"name": ptag["name"]})
        return self._make_ptag_dict(ptag_db)

    def _make_ptag_dict(self, ptag, fields=None):
        ptag_dict = {"id": ptag.id,
                    "name": ptag.name,
                    "tag_type": ptag.tag_type,
                    "tag_id": ptag.tag_id,
                    "floatingip_id": ptag.floatingip_id,
                    "floating_ip_address": ptag.floatingip_address,
                    "router_id": ptag.router_id,
                    "tenant_id": ptag.tenant_id}
        return self._fields(ptag_dict, fields)

    def _check_floatingip_in_use(self, context, fp_id):
        query = context.session.query(PolicyTag)
        try:
            ptag = query.filter_by(floatingip_id=fp_id).one()
            if ptag:
                raise p_excep.FloatingIPAlreadyInUse(id=fp_id)
        except exc.NoResultFound:
            pass



# Copyright 2015 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import sys

import mock

from networking_plumgrid.neutron.tests.unit.cli import test_cli20
from networking_plumgrid.neutronclient.l2gw_client_ext import (
    _l2_gateway_connection as l2_gateway_connection)

from neutronclient import shell


class CLITestV20ExtensionL2GWJSON(test_cli20.CLITestV20Base):
    def setUp(self):
        # need to mock before super because extensions loaded on instantiation
        self._mock_extension_loading()
        super(CLITestV20ExtensionL2GWJSON, self).setUp(plurals={'tags': 'tag'})

    def _mock_extension_loading(self):
        ext_pkg = 'neutronclient.common.extension'
        contrib = mock.patch(ext_pkg + '._discover_via_entry_points').start()
        contrib.return_value = [("_l2_gateway_connection",
                                 l2_gateway_connection)]
        return contrib

    def test_ext_cmd_loaded(self):
        """Tests l2gw-connection commands loaded."""
        shell.NeutronShell('2.0')
        ext_cmd = {'l2-gateway-connection-list':
                   l2_gateway_connection.L2GatewayConnectionList,
                   'l2-gateway-connection-create':
                   l2_gateway_connection.L2GatewayConnectionCreate,
                   'l2-gateway-connection-delete':
                   l2_gateway_connection.L2GatewayConnectionDelete,
                   'l2-gateway-connection-show':
                   l2_gateway_connection.L2GatewayConnectionShow}
        self.assertDictContainsSubset(ext_cmd, shell.COMMANDS['2.0'])

    def test_create_l2gateway_connection(self):
        """Test Create l2gateway-connection."""

        resource = 'l2_gateway_connection'
        cmd = l2_gateway_connection.L2GatewayConnectionCreate(
            test_cli20.MyApp(sys.stdout), None)
        l2gw_id = 'l2gw-id'
        net_id = 'net-id'
        args = [l2gw_id, net_id, '--segmentation-id', 'seg-id']
        position_names = ['l2_gateway_id', 'network_id', 'segmentation_id']
        position_values = [l2gw_id, net_id, 'seg-id']
        self._test_create_resource(resource, cmd, l2gw_id, 'myid', args,
                                   position_names, position_values)

    def test_list_l2gateway_connection(self):
        """Test List l2gateway-connections."""

        resources = "l2_gateway_connections"
        cmd = l2_gateway_connection.L2GatewayConnectionList(
            test_cli20.MyApp(sys.stdout), None)
        self._test_list_resources(resources, cmd, True)

    def test_delete_l2gateway_connection(self):
        """Test Delete l2gateway-connection."""

        resource = 'l2_gateway_connection'
        cmd = l2_gateway_connection.L2GatewayConnectionDelete(
            test_cli20.MyApp(sys.stdout), None)
        my_id = 'my-id'
        args = [my_id]
        self._test_delete_resource(resource, cmd, my_id, args)

    def test_show_l2gateway_connection(self):
        """Test Show l2gateway-connection: --fields id --fields name myid."""

        resource = 'l2_gateway_connection'
        cmd = l2_gateway_connection.L2GatewayConnectionShow(
            test_cli20.MyApp(sys.stdout), None)
        args = ['--fields', 'id', '--fields', 'name', self.test_id]
        self._test_show_resource(resource, cmd, self.test_id, args,
                                 ['id', 'name'])



# Copyright 2015 PLUMgrid, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
PLUMgrid plugin physical attachment point extension unit tests
"""

from networking_plumgrid.neutron.plugins.common import \
    exceptions as plum_excep
from networking_plumgrid.neutron.plugins.extensions import \
    physical_attachment_point as ext_pap
from networking_plumgrid.neutron.tests.unit import \
    test_networking_plumgrid as test_pg
from neutron.api import extensions
from neutron import context
from neutron import manager
from neutron.tests.unit.api import test_extensions as test_ext
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class PAPExtensionManager(object):

    def get_resources(self):
        return ext_pap.Physical_attachment_point.get_resources()

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class PhysicalAttachmentPointTestCase(test_pg.PLUMgridPluginV2TestCase):
    def setUp(self, plugin=None, ext_mgr=None):
        super(PhysicalAttachmentPointTestCase, self).setUp()
        ext_mgr = PAPExtensionManager()
        extensions.PluginAwareExtensionManager._instance = None
        self.ext_api = test_ext.setup_extensions_middleware(ext_mgr)


class TestPhysicalAttachmentPoint(PhysicalAttachmentPointTestCase):
    def test_create_pap(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_without_td(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=None)
        LOG.error(pap)
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_with_td_name(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["name"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_lacp_true_hash_L2(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, lacp=True,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_as_false_lacp_false(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, active_standby=False,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_as_false_lacp_true(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, active_standby=False,
                                  lacp=True, transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_as_true_lacp_false(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, active_standby=True,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_as_true_lacp_true(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, active_standby=True,
                                  lacp=True, transit_domain_id=td["id"])
        self.assertRaises(plum_excep.PLUMgridException,
            plugin.create_physical_attachment_point,
            admin_context, pap)

    def test_create_pap_lacp_false_hash_not_L2(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces, lacp='False',
                  hash_mode="L3", transit_domain_id=td["id"])
        self.assertRaises(plum_excep.PLUMgridException,
            plugin.create_physical_attachment_point,
            admin_context, pap)

    def test_create_pap_no_interfaces(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_multi_interfaces(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "test_host1", "interface": "ifc1"},
                      {"hostname": "test_host1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_pap_max_hosts(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h%d" % i, "interface": "ifc%d" % j}
                      for i in range(1, 16) for j in range(1, 2)]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap["physical_attachment_point"]["id"] = pap_ret["id"]
        self.assertEqual(pap_ret, pap["physical_attachment_point"])

    def test_create_update_show_pap01(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(transit_domain_id=td["id"])
        pap_old = plugin.create_physical_attachment_point(
                      admin_context, pap)
        add_interfaces = [{"hostname": "test_host", "interface": "ifc"}]
        pap_new = self._make_pap_update_dict(add_interfaces=add_interfaces,
                                             transit_domain_id=td["id"])

        pap_new_ret = plugin.update_physical_attachment_point(
                              admin_context, pap_old["id"], pap_new)
        pap_new_get = plugin.get_physical_attachment_point(
                              admin_context, pap_old["id"])

        pap["physical_attachment_point"]["id"] = pap_old["id"]
        self.assertEqualUpdate(pap_old, pap["physical_attachment_point"])
        self.assertEqualUpdate(pap_new_ret, pap_new_get)

    def test_create_update_show_pap02(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"},
                      {"hostname": "h2", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_old = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap_new = self._make_pap_update_dict(remove_interfaces=interfaces,
                                             transit_domain_id=td["id"])
        pap_new_ret = plugin.update_physical_attachment_point(
                              admin_context, pap_old["id"], pap_new)
        pap_new_get = plugin.get_physical_attachment_point(
                              admin_context, pap_old["id"])
        pap["physical_attachment_point"]["id"] = pap_old["id"]
        self.assertEqualUpdate(pap_old, pap["physical_attachment_point"])
        self.assertEqualUpdate(pap_new_ret, pap_new_get)

    def test_create_update_show_pap03(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_old = plugin.create_physical_attachment_point(
                      admin_context, pap)

        add_interfaces = [{"hostname": "h1", "interface": "ifc2"},
                      {"hostname": "h2", "interface": "ifc1"}]
        pap_new = self._make_pap_update_dict(add_interfaces=add_interfaces,
                                             transit_domain_id=td["id"])
        pap_new_ret = plugin.update_physical_attachment_point(
                              admin_context, pap_old["id"], pap_new)
        pap_new_get = plugin.get_physical_attachment_point(
                              admin_context, pap_old["id"])
        pap["physical_attachment_point"]["id"] = pap_old["id"]
        self.assertEqual(pap_old, pap["physical_attachment_point"])
        self.assertEqualUpdate(pap_new_ret, pap_new_get)

    def test_create_show_pap_by_uuid(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"},
                      {"hostname": "h2", "interface": "ifc2"},
                      {"hostname": "h2", "interface": "ifc2"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap_get_ret = plugin.get_physical_attachment_point(
                          admin_context, pap_ret["id"])
        self.assertEqual(pap_get_ret, pap_ret)

    def test_create_show_pap_by_uuid_no_ifcs(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        pap_get_ret = plugin.get_physical_attachment_point(
                          admin_context, pap_ret["id"])
        self.assertEqual(pap_get_ret, pap_ret)

    def test_create_delete_pap_multi_ifcs(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()
        interfaces = [{"hostname": "h1", "interface": "ifc1"},
                      {"hostname": "h2", "interface": "ifc2"},
                      {"hostname": "h2", "interface": "ifc2"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        plugin.delete_physical_attachment_point(
                          admin_context, pap_ret["id"])

    def test_create_delete_pap_no_ifcs(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(transit_domain_id=td["id"])
        pap_ret = plugin.create_physical_attachment_point(
                      admin_context, pap)
        plugin.delete_physical_attachment_point(
                          admin_context, pap_ret["id"])

    def test_create_update_get_delete_pap01(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"},
                      {"hostname": "h2", "interface": "ifc1"}]
        td = self._create_transit_domain(admin_context, plugin)
        pap = self._make_pap_dict(interfaces=interfaces,
                                  transit_domain_id=td["id"])
        pap_old = plugin.create_physical_attachment_point(
                      admin_context, pap)

        add_interfaces = [{"hostname": "h2", "interface": "ifc2"}]
        pap_new = self._make_pap_update_dict(add_interfaces=add_interfaces,
                                             transit_domain_id=td["id"])
        pap_new_ret = plugin.update_physical_attachment_point(
                              admin_context, pap_old["id"], pap_new)
        pap_get_ret = plugin.get_physical_attachment_point(
                          admin_context, pap_new_ret["id"])
        self.assertEqualUpdate(pap_get_ret, pap_new_ret)
        plugin.delete_physical_attachment_point(
            admin_context, pap_get_ret["id"])

    def test_get_all(self):
        plugin = manager.NeutronManager.get_plugin()
        admin_context = context.get_admin_context()

        interfaces = [{"hostname": "h1", "interface": "ifc1"},
                      {"hostname": "h2", "interface": "ifc2"}]
        td1 = self._create_transit_domain(admin_context, plugin)
        pap_1 = self._make_pap_dict(interfaces=interfaces,
                                    transit_domain_id=td1["id"])
        pap_1_ret = plugin.create_physical_attachment_point(
                      admin_context, pap_1)

        interfaces = [{"hostname": "h1", "interface": "ifc2"},
                      {"hostname": "h1", "interface": "ifc3"},
                      {"hostname": "h1", "interface": "ifc4"}]
        td2 = self._create_transit_domain(admin_context, plugin)
        pap_2 = self._make_pap_dict(interfaces=interfaces,
                                    transit_domain_id=td2["id"])
        pap_2_ret = plugin.create_physical_attachment_point(
                              admin_context, pap_2)
        pap_list_get = plugin.get_physical_attachment_points(
                          admin_context)
        cmp(pap_list_get, [pap_1_ret, pap_2_ret])

    def _make_pap_dict(self, lacp=False, active_standby=False,
                       hash_mode="L2",
                       transit_domain_id="test_id",
                       interfaces=[]):
        return {"physical_attachment_point": {
                   "tenant_id": "test_tenant",
                   "name": "test_name",
                   "hash_mode": hash_mode,
                   "lacp": lacp,
                   "active_standby": active_standby,
                   "implicit": False,
                   "transit_domain_id": transit_domain_id,
                   "interfaces": interfaces}}

    def _make_pap_update_dict(self, lacp=False, active_standby=False,
                              hash_mode="L2",
                              transit_domain_id="test_id",
                              add_interfaces=[], remove_interfaces=[]):
        return {"physical_attachment_point": {
                   "tenant_id": "test_tenant",
                   "name": "test_name",
                   "hash_mode": hash_mode,
                   "lacp": lacp,
                   "active_standby": active_standby,
                   "implicit": False,
                   "transit_domain_id": transit_domain_id,
                   "add_interfaces": add_interfaces,
                   "remove_interfaces": remove_interfaces}}

    def _create_transit_domain(self, admin_context, plugin):
        td = {"transit_domain": {
                  "tenant_id": "test_tenant",
                  "name": "td",
                  "implicit": False}}
        res = plugin.create_transit_domain(admin_context, td)
        return res

    def _update_interfaces(self, **kwargs):
        interfaces = kwargs.pop("interfaces", [])
        add_interfaces = kwargs.pop("add_interfaces", [])
        remove_interfaces = kwargs.pop("remove_interfaces", [])

        interfaces.extend(add_interfaces)
        for interface in remove_interfaces:
            interfaces = interfaces.remove(interface)

        if not interfaces:
            interfaces = []
        return interfaces

    def assertEqualUpdate(self, pap_ref, pap_actual):
        def make_pap_dict(pap_db):
            if "add_interfaces" in pap_db:
                del pap_db["add_interfaces"]
            if "remove_interfaces" in pap_db:
                del pap_db["remove_interfaces"]
            return pap_db

        pap_ref = make_pap_dict(pap_ref)
        pap_actual = make_pap_dict(pap_actual)
        return self.assertEqual(pap_ref, pap_actual)



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from networking_plumgrid._i18n import _
from neutronclient.common import extension
from neutronclient.common import utils


class Endpoint(extension.NeutronClientExtension):
    resource = 'endpoint'
    resource_plural = 'endpoints'
    path = 'endpoints'
    object_path = '/%s' % path
    resource_path = '/%s/%%s' % path
    versions = ['2.0']


def args2body(self, parsed_args):
    try:
        if parsed_args.ep_groups:
            epg_list = parsed_args.ep_groups
        else:
            epg_list = []
        epg_dict = []
        for epg in epg_list:
            if "id" in epg:
                epg = {'id': epg['id']}
            elif "name" in epg:
                epg = {'id': epg['name']}
            else:
                raise KeyError("ID for Endpoint Group is required.")
            epg_dict.append(epg)
        if parsed_args.ip_port_mask:
            ip_port_list = parsed_args.ip_port_mask
        else:
            ip_port_list = []
        ip_port_mask = None
        if len(ip_port_list) > 1:
            raise Exception("Cannot use multiple ip-port classifcation "
                            "criteria. Please specify only one.")
        for ip_port in ip_port_list:
            if "ip" not in ip_port or \
               "port" not in ip_port or \
               "mask" not in ip_port:
                raise Exception("IP Address, Port and Port Mask need to be "
                                "specifed when using --ip-port classification")
            else:
                ip_port_mask = ip_port["ip"] + "_" + ip_port["port"] \
                               + "_" + ip_port["mask"]

        if ((parsed_args.port_id or parsed_args.ip_mask or
             parsed_args.ip_port_mask or parsed_args.label) and not
            parsed_args.ep_groups):
            raise Exception("Atleast one endpoint group(--endpoint-group)"
                            " is required for association.")
        if (parsed_args.ep_groups and (not parsed_args.port_id and not
            parsed_args.ip_mask and not parsed_args.ip_port_mask and
            not parsed_args.label)):
            raise Exception("Please specify an association criteria. "
                            "Supported criterion are: ['--port', "
                            "'--ip-mask', '--ip-port']")
        if ((parsed_args.ip_port_mask and parsed_args.ip_mask) or
            (parsed_args.ip_port_mask and parsed_args.port_id) or
            (parsed_args.ip_port_mask and parsed_args.label) or
            (parsed_args.ip_mask and parsed_args.port_id) or
            (parsed_args.ip_mask and parsed_args.label) or
            (parsed_args.label and parsed_args.port_id)):
            raise Exception("Multiple association criterion for endpoint "
                            "specified. Please specify only one criteria "
                            "per endpoint.")
        if (not parsed_args.port_id and not parsed_args.ip_mask and
            not parsed_args.ip_port_mask and not parsed_args.label):
            raise Exception("Please specify an association criteria. "
                            "Supported criterion are: ['--port', "
                            "'--ip-mask', '--ip-port']")
        if parsed_args.name:
            ep_name = parsed_args.name
            body = {'endpoint': {'name': ep_name}}
        else:
            body = {'endpoint': {}}
        if parsed_args.ep_groups:
            body['endpoint']['ep_groups'] = epg_dict
        if parsed_args.ip_port_mask:
            body['endpoint']['ip_port'] = ip_port_mask
        if parsed_args.ip_mask:
            body['endpoint']['ip_mask'] = parsed_args.ip_mask
        if parsed_args.port_id:
            body['endpoint']['port_id'] = parsed_args.port_id
        if parsed_args.label:
            body['endpoint']['label'] = parsed_args.label
        return body
    except KeyError as err:
        raise Exception("KeyError: " + str(err))


class EndpointCreate(extension.ClientExtensionCreate,
                    Endpoint):
    """Create a Endpoint."""

    shell_command = 'endpoint-create'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name', metavar='<ENDPOINT-NAME>',
            help=_('Descriptive name for endpoint.'))
        parser.add_argument('--port', dest='port_id',
                        metavar='PORT UUID/NAME',
                        help=_('Port UUID/NAME'))
        parser.add_argument('--ip-mask', dest='ip_mask',
                        help=_('IP Address/Mask'))
        parser.add_argument(
                   '--ip-port',
                   metavar='ip=IP-ADDRESS, port=PORT-NUMBER, mask=PORT-MASK',
                   action='append', dest='ip_port_mask', type=utils.str2dict,
                   help=_('IP Address and Port Mask to specify port range'))
        parser.add_argument('--label', dest='label',
                        help=_('Label/Context of network function.'))
        parser.add_argument(
                   '--endpoint-group',
                   metavar='id=ENPOINT-GROUP-UUID, name=ENDPOINT-GROUP-NAME',
                   action='append', dest='ep_groups', type=utils.str2dict,
                   help=_('Endpoint Groups to be associated with endpoint '
                          'id=<endpoint-group-uuid>,name=<endpoint-group-name>'
                          ' (--endpoint-group option can be repeated)'))

    def args2body(self, parsed_args):
        body = args2body(self, parsed_args)
        if parsed_args.tenant_id:
            (body['endpoint']
                 ['tenant_id']) = parsed_args.tenant_id
        return body


class EndpointList(extension.ClientExtensionList,
                   Endpoint):
    """List endpoints that belong to a given tenant."""

    shell_command = 'endpoint-list'
    list_columns = ['id', 'name', 'ep_groups', 'port_id', 'ip_mask',
                    'ip_port', 'label']
    pagination_support = True
    sorting_support = True


class EndpointShow(extension.ClientExtensionShow,
                   Endpoint):
    """Show information of a given endpoints."""

    shell_command = 'endpoint-show'


class EndpointDelete(extension.ClientExtensionDelete,
                     Endpoint):
    """Delete a given endpoint."""

    shell_command = 'endpoint-delete'


class EndpointUpdate(extension.ClientExtensionUpdate,
                     Endpoint):
    """Update a given endpoint."""

    shell_command = 'endpoint-update'

    def add_known_arguments(self, parser):
        parser.add_argument(
            '--name', metavar='name',
            help=_('Descriptive name for Endpoint'))
        parser.add_argument(
            '--add-endpoint-group',
            metavar='id=ENPOINT-GROUP-UUID, name=ENDPOINT-GROUP-NAME',
            action='append', dest='add_endpoint_groups', type=utils.str2dict,
            help=_('Endpoint Group ID to be associated with this Endpoint.'
                   '(--add-endpoint-group option can be repeated)'))
        parser.add_argument(
            '--remove-endpoint-group',
            metavar='id=ENPOINT-GROUP-UUID, name=ENDPOINT-GROUP-NAME',
            action='append', dest='remove_endpoint_groups',
            type=utils.str2dict,
            help=_('Endpoint Group ID to be removed from this Endpoint.'
                   '(--remove-endpoint-group option can be repeated)'))

    def args2body(self, parsed_args):
        try:
            if parsed_args.add_endpoint_groups:
                add_endpoint_groups = parsed_args.add_endpoint_groups
            else:
                add_endpoint_groups = []
            add_endpoint_group_dict = []
            for epg in add_endpoint_groups:
                if "id" in epg:
                    epg = {'id': epg['id']}
                elif "name" in epg:
                    epg = {'id': epg["name"]}
                else:
                    raise KeyError("ID for Endpoint Group is required.")
                add_endpoint_group_dict.append(epg)
            if parsed_args.remove_endpoint_groups:
                remove_endpoint_groups = parsed_args.remove_endpoint_groups
            else:
                remove_endpoint_groups = []
            remove_endpoint_group_dict = []
            for epg in remove_endpoint_groups:
                if "id" in epg:
                    epg = {'id': epg['id']}
                elif "name" in epg:
                    epg = {'id': epg['name']}
                else:
                    raise KeyError("ID for Endpoint Group is required.")
                remove_endpoint_group_dict.append(epg)

            if (parsed_args.add_endpoint_groups and
                parsed_args.remove_endpoint_groups):
                if len(set([tuple(d.items()) for d in
                       add_endpoint_group_dict]).
                       intersection(set([tuple(d.items()) for d
                       in remove_endpoint_group_dict]))) > 0:
                    raise Exception("Duplicate endpoint groups found.")
            if parsed_args.name:
                ep_name = parsed_args.name
                body = {'endpoint': {'name': ep_name}}
            else:
                body = {'endpoint': {}}
            if parsed_args.add_endpoint_groups:
                (body['endpoint'][
                 'add_endpoint_groups']) = add_endpoint_group_dict
            if parsed_args.remove_endpoint_groups:
                (body['endpoint'][
                 'remove_endpoint_groups']) = remove_endpoint_group_dict
            return body
        except KeyError as err:
            raise Exception("KeyError: " + str(err))



# Copyright 2016 PLUMgrid, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from networking_plumgrid.neutron.plugins.common import constants as \
    const
from networking_plumgrid.neutron.plugins.common import \
    policy_exceptions as p_excep
from networking_plumgrid.neutron.plugins.extensions import \
    policyrule as ext_pr
from random import randint


def _validate_policy_rule_config(pr_db):
    """
    Helper function to validate policy rule config
    """
    if "protocol" in pr_db:
        if not _validate_policy_rule_protocol(pr_db):
            raise p_excep.UnsupportedProtocol(protocol=pr_db["protocol"])
        if pr_db["protocol"] == "any":
            if (("src_port_range" in pr_db or "dst_port_range" in pr_db)
                and (pr_db["src_port_range"] or pr_db["dst_port_range"])):
                raise p_excep.ProtocolNotSpecified()
    if "action" in pr_db:
        if not _validate_policy_rule_action(pr_db):
            raise p_excep.UnsupportedAction(action=pr_db["action"])
    if "src_port_range" in pr_db:
        _validate_port_range(pr_db["src_port_range"])
    if "dst_port_range" in pr_db:
        _validate_port_range(pr_db["dst_port_range"])
    if _validate_port_range_for_icmp(pr_db):
        return True
    return True


def _validate_policy_rule_protocol(pr_db):
    """
    Helper function to validate protocol for policy rule
    """
    if pr_db["protocol"] in const.SUPPORTED_POLICY_RULE_PROTOCOLS:
        return True
    return False


def _validate_policy_rule_action(pr_db):
    """
    Helper function to validate action for policy rule
    """
    if pr_db["action"] in const.SUPPORTED_POLICY_RULE_ACTIONS:
        if (("action_target" not in pr_db or not pr_db["action_target"])
            and pr_db["action"] == 'copy'):
            raise p_excep.InvalidInputActionTarget()
        if (pr_db["action"] == 'allow' and ("action_target" in pr_db
            and pr_db["action_target"])):
            raise p_excep.InvalidInputActionTarget()
        return True
    return False


def _validate_port_range(port_range):
    if port_range is None:
        return port_range
    try:
        lower_bound, upper_bound = port_range.split('-')
        lower_bound_val = int(lower_bound)
        upper_bound_val = int(upper_bound)
    except (ValueError, TypeError):
        port_range_min = randint(1, 65535)
        port_range_max = randint(port_range_min, 65535)
        raise ext_pr.PolicyRuleInvalidPortRange(port=port_range,
                                             min=str(port_range_min),
                                             max=str(port_range_max))

    if ((lower_bound_val >= 0 and lower_bound_val <= 65535 and
        upper_bound_val >= 0 and upper_bound_val <= 65535) and
        lower_bound_val <= upper_bound_val):
        return port_range
    else:
        port_range_min = randint(1, 65535)
        port_range_max = randint(port_range_min, 65535)
        raise ext_pr.PolicyRuleInvalidPortRange(port=port_range,
                                             min=str(port_range_min),
                                             max=str(port_range_max))


def _validate_port_range_for_icmp(pr_db):
    if ("protocol" in pr_db and
        pr_db["protocol"] == const.SUPPORTED_POLICY_RULE_PROTOCOLS[3]):
        if ("src_port_range" in pr_db and
            pr_db["src_port_range"] is not None):
            raise p_excep.NotSupportedPortRangeICMP()
        elif ("dst_port_range" in pr_db and
              pr_db["dst_port_range"] is not None):
            raise p_excep.NotSupportedPortRangeICMP()
    return True
